import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.util.*;
import java.io.*;

public class TSTDemo extends Applet {
	private TernarySearchTree tst = TSTLoader.getTST();
	private TextListenerAction lookupAction = new LookupAction(), noAction = new NoAction(), preAction = new PreAction(), spellAction = new SpellAction();
	private TextListenerAction currentAction = lookupAction;
	private InfoPanel infoPanel;
	private TextField textField;
	private TextArea textArea;
	
	public static void main(String args[]) {
		DemoFrame app = new DemoFrame("Ternary Search Tree Demo");
		app.setSize(500,600);
		app.show();
	}
         
	public void init() {
		Font defaultFont = new Font("SansSerif", Font.BOLD, 12);
		setFont(defaultFont);
		FontMetrics fontMetrics = getFontMetrics(defaultFont);
		
		// checkboxes
		CheckboxGroup cbg = new CheckboxGroup();
		Checkbox lookupCheck = new Checkbox("lookup", cbg, true);
		Checkbox addCheck = new Checkbox("add", cbg, false);
		Checkbox preCheck = new Checkbox("match prefix", cbg, false);
		Checkbox spellCheck = new Checkbox("check spelling    ", cbg, false);
		Panel checkboxPanel = new Panel();
		checkboxPanel.setLayout(new GridLayout(2, 2));
		checkboxPanel.add(lookupCheck);
		checkboxPanel.add(preCheck);
		checkboxPanel.add(addCheck);
		checkboxPanel.add(spellCheck);
		
		// text field and label
		FixedSizeLabel textFieldLabel = new FixedSizeLabel("word", fontMetrics);
		textField = new TextField(10);

		// text area and label
		FixedSizeLabel textAreaLabel = new FixedSizeLabel("definition", fontMetrics);
		textArea = new TextArea(10, 10);		
		
		// text field and text area panel
		final Panel displayPanel = new Panel();
		GridBagLayout gridbag = new GridBagLayout();
		GridBagConstraints c = new GridBagConstraints();
		displayPanel.setLayout(gridbag);
		Insets topLeftInsets = new Insets(6, 0, 6, 4);
		c.insets = topLeftInsets;
		c.fill = GridBagConstraints.NONE;
		c.anchor = GridBagConstraints.NORTHEAST;
		c.weightx = 0;
		c.weighty = 0;
		displayPanel.add(textFieldLabel, c);
		c.gridx = 1;
		c.weightx = 1;
		c.fill = GridBagConstraints.HORIZONTAL;
		Insets topRightInsets = new Insets(6, 0, 6, 0);
		c.insets = topRightInsets;
		displayPanel.add(textField, c);
		c.insets = new Insets(6, 4, 0, 4);
		c.weightx = 0;
		c.weighty = 1;
		c.gridx = 0;
		c.gridy = 1;
		c.fill = GridBagConstraints.NONE;
		c.anchor = GridBagConstraints.NORTHEAST;
		Insets bottomLeftInsets = new Insets(6, 0, 0, 4);
		c.insets = bottomLeftInsets;
		displayPanel.add(textAreaLabel, c);
		c.weightx = 1;
		c.gridx = 1;
		c.fill = GridBagConstraints.BOTH;
		Insets bottomRightInsets = new Insets(6, 0, 0, 0);
		c.insets = bottomRightInsets;
		displayPanel.add(textArea, c);
		
		// button panel
		final Panel buttonPanel = new Panel();
		Panel blankPanel = new Panel();
		Panel addPanel = new Panel();
		Button addButton = new Button("add");
		SpacerLabel spacer = new SpacerLabel(textAreaLabel.getMinimumSize().width);
		addPanel.add(spacer);
		addPanel.add(addButton);
		final CardLayout cardLayout = new CardLayout();
		buttonPanel.setLayout(cardLayout);
		buttonPanel.add(blankPanel, "blank");
		buttonPanel.add(addPanel, "add");
		cardLayout.show(buttonPanel, "blank");
		

		// info panel
		infoPanel = new InfoPanel(fontMetrics);
		setLayout(new GridBagLayout());
		Insets zeroInsets = new Insets(0, 0, 0, 0);
		c.insets = zeroInsets;
		c.gridx = c.gridy = 0;
		c.weighty = 0;
		c.weightx = 1;
		c.fill = GridBagConstraints.NONE;
		c.anchor = GridBagConstraints.NORTHEAST;
		add(checkboxPanel, c);
		c.fill = GridBagConstraints.BOTH;
		c.gridy = 1;
		c.weighty = 1;
		c.anchor = GridBagConstraints.CENTER;
		add(displayPanel, c);
		c.gridy = 2;
		c.weighty = 0;
		c.anchor = GridBagConstraints.CENTER;
		c.fill = GridBagConstraints.NONE;
		c.insets = topRightInsets;
		add(buttonPanel, c);
		c.gridy = 3;
		c.weighty = 0;
		c.anchor = GridBagConstraints.EAST;
		c.fill = GridBagConstraints.NONE;
		c.insets = zeroInsets;
		add(infoPanel, c);
		
		// add listeners to checkboxes
		lookupCheck.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent itemEvent){
				if(itemEvent.getStateChange() == ItemEvent.SELECTED) {
					textField.setText("");
					textArea.setText("");
					currentAction = lookupAction;
					cardLayout.show(buttonPanel, "blank");
				}
			}
		});
		addCheck.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent itemEvent){
				if(itemEvent.getStateChange() == ItemEvent.SELECTED) {
					textField.setText("");
					textArea.setText("");
					cardLayout.show(buttonPanel, "add");
					currentAction = noAction;
				}
			}
		});
		preCheck.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent itemEvent){
				if(itemEvent.getStateChange() == ItemEvent.SELECTED) {
					textField.setText("");
					textArea.setText("");
					currentAction = preAction;
					cardLayout.show(buttonPanel, "blank");
				}
			}
		});
		spellCheck.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent itemEvent){
				if(itemEvent.getStateChange() == ItemEvent.SELECTED) {
					textField.setText("");
					textArea.setText("");
					currentAction = spellAction;
					cardLayout.show(buttonPanel, "blank");
				}
			}
		});
		
		// add text listener to query box
		textField.addTextListener(new TextListener() {
			public void textValueChanged(TextEvent textEvent) {
				String textFieldContents = textField.getText();
				if(textFieldContents.equals("")) {
					textArea.setText("");
					infoPanel.setOpValues(0);	
					infoPanel.setOpTime(0);
					infoPanel.repaint();
					return;
				}
				currentAction.doAction(textFieldContents);
			}
		});
		
		// add button listener
		addButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent event) {
				String textFieldContents = textField.getText();
				String textAreaContents = textArea.getText();
				if(textFieldContents.equals("") || textAreaContents.equals("")) return;
				long time = System.currentTimeMillis();
				tst.put(textFieldContents, textAreaContents);
				time = System.currentTimeMillis() - time;
				infoPanel.setNumWords(tst.numDataNodes());
				infoPanel.setOpTime(time);
				infoPanel.setOpValues(0);
				textField.setText("");
				textArea.setText("");
				infoPanel.repaint();
			}
		});

		infoPanel.setNumWords(tst.numDataNodes());
		setBackground(Color.white);

	}
	
	public void start() {
	}
	
	public void stop() {
	}
	
	public void destroy() {
	}
	
	private class FixedSizeLabel extends Label {
		private Dimension minSize;
		public FixedSizeLabel(String initialText, FontMetrics fm) {
			super(initialText, Label.RIGHT);
			minSize = new Dimension(fm.stringWidth("definition") + 4, fm.getMaxAscent() + fm.getMaxDescent() + fm.getLeading());
		}
		public Dimension getMinimumSize() {return minSize;}
		public Dimension getPreferredSize() {return minSize;}
		public Dimension getMaximumSize() {return minSize;}
	}
	
	private class SpacerLabel extends Label {
		private Dimension minSize;
		public SpacerLabel(int width) {
			minSize = new Dimension(width, 0);
		}
		public Dimension getMinimumSize() {return minSize;}
		public Dimension getPreferredSize() {return minSize;}
		public Dimension getMaximumSize() {return minSize;}
	}
	
	private class InfoPanel extends Panel {
		private int lineHeight, leading, valueXCoordinate;
		private String[] messagePrefixes = {"Number of words in dictionary: ", "Time for last operation (milliseconds): ", "Values returned by last operation: "};
		private long[] messageValues = new long[messagePrefixes.length];
		private Dimension minSize;
		public InfoPanel(FontMetrics fm) {
			this.lineHeight = fm.getMaxAscent() + fm.getMaxDescent();
			this.leading = fm.getLeading();
			int maxLength = 0;
			for(int i = 0; i < messagePrefixes.length; i++) maxLength = Math.max(maxLength, fm.stringWidth(messagePrefixes[i]));
			minSize = new Dimension(maxLength + fm.stringWidth("500000") + 4, (leading * (messagePrefixes.length + 1)) + (lineHeight * messagePrefixes.length) + 12);
			valueXCoordinate = minSize.width - fm.stringWidth("500000");
		}
		public void paint(Graphics g) {
			int currentY = leading + lineHeight + 6;
			for(int i = 0; i < messagePrefixes.length; i++) {
				g.drawString(messagePrefixes[i], 0, currentY);
				g.drawString(String.valueOf(messageValues[i]), valueXCoordinate, currentY);
				currentY += lineHeight + leading;
			}
		}
		public Dimension getMinimumSize() {
			return minSize;
		}
		public Dimension getPreferredSize() {
			return minSize;
		}
		public Dimension getMaximumSize() {
			return minSize;
		}
		public void setNumWords(int num) {messageValues[0] = num;}
		public void setOpTime(long time) {messageValues[1] = time;}
		public void setOpValues(int numValues) {messageValues[2] = numValues;}
	}
	
	private abstract class TextListenerAction {
		public abstract void doAction(String textFieldContents);
	}
	
	private final class LookupAction extends TextListenerAction {
		public final void doAction(String textFieldContents) {
			long time = System.currentTimeMillis();
			String definition = (String)tst.get(textFieldContents);
			time = System.currentTimeMillis() - time;
			textArea.setText(definition);
			infoPanel.setOpValues(1);
			infoPanel.setOpTime(time);
			infoPanel.repaint();
		}
	}
	
	private final class NoAction extends TextListenerAction {
		public final void doAction(String textFieldContents) {
		}
	}
	
	private final class PreAction extends TextListenerAction {
		public final void doAction(String textFieldContents) {
			long time = System.currentTimeMillis();
			String matchingKeys = tst.matchPrefixString(textFieldContents);
			time = System.currentTimeMillis() - time;
			textArea.setText(matchingKeys);
			infoPanel.setOpValues(tst.getLastNumReturnValues());	
			infoPanel.setOpTime(time);
			infoPanel.repaint();
		}
	}
	
	private final class SpellAction extends TextListenerAction {
		public final void doAction(String textFieldContents) {
			int returnValues = 0;
			boolean[] hasValue = {false, false, false, false};
			int currentLastNum;
			long time = System.currentTimeMillis();
			tst.setMatchAlmostDiff(0);
			String[] intermediateResults = {"", "", "", ""};
			intermediateResults[0] = tst.matchAlmostString(textFieldContents);
			currentLastNum = tst.getLastNumReturnValues();
			if(currentLastNum > 0) hasValue[0] = true;
			returnValues += currentLastNum;
			tst.setMatchAlmostDiff(1);
			intermediateResults[1] = tst.matchAlmostString(textFieldContents);
			currentLastNum = tst.getLastNumReturnValues();
			if(currentLastNum > 0) hasValue[1] = true;
			returnValues += currentLastNum;
			
			TernarySearchTree uniqueKeyTree;
			if(textFieldContents.length() > 3) {
				String currentResult;
				uniqueKeyTree = new TernarySearchTree();
				Object keyMarker = new Object();
				String altered;
				for(int i = 0; i <= textFieldContents.length(); i++) {
					altered = textFieldContents.substring(0, i) + "*" + textFieldContents.substring(i);
					currentResult = tst.matchAlmostString(altered);
					System.out.println("");
					System.out.println("altered string: " + altered);
					System.out.println("curent result: " + currentResult);
					
					if(currentResult.length() > 0) uniqueKeyTree.put(currentResult, keyMarker);
				}
				
				tst.setMatchAlmostDiff(0);
				char holdChar;
				for(int i = 0; i < textFieldContents.length(); i++) {
					holdChar = textFieldContents.charAt(i);
					altered = textFieldContents.substring(0, i) + textFieldContents.substring(i + 1);
					currentResult = tst.matchAlmostString(altered);
					if(currentResult.length() > 0) uniqueKeyTree.put(currentResult, keyMarker);
				}
				intermediateResults[2] = uniqueKeyTree.sortKeysString();
				if(intermediateResults[2].length() > 0) hasValue[2] = true;
				returnValues += uniqueKeyTree.numDataNodes();
			}
			if(textFieldContents.length() > 5) {
				tst.setMatchAlmostDiff(2);
				intermediateResults[3] = tst.matchAlmostString(textFieldContents);
				currentLastNum = tst.getLastNumReturnValues();
				if(currentLastNum > 0) hasValue[3] = true;
				returnValues += currentLastNum;
			}

			String finalResult = (hasValue[0]) ? intermediateResults[0] : "";
			for(int i = 0; i < 3; i++) {
				if(hasValue[i] && hasValue[i]) finalResult += "\n";
				finalResult += intermediateResults[i+1];
			}
			time = System.currentTimeMillis() - time;
			textArea.setText(finalResult);
			infoPanel.setOpTime(time);
			infoPanel.setOpValues(returnValues);
			infoPanel.repaint();
		}
	}
}

class DemoFrame extends Frame {
	public DemoFrame(String frameTitle) {
		super(frameTitle);

		TSTDemo applet = new TSTDemo();
		applet.init();
		applet.start();
		add(applet, "Center");

		addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent event) {
				dispose();
				System.exit(0);
			}	
		});
	}
}
